/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package ants;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

import support.AntRunner;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;

public class AntSystem {

	private static List<Ant> ants;
	private static List<Ant> tba;
	private static List<Ant> tbr;
	
	public static Iterator<Ant> iterator;

	private static List<AntRunner> runners;

	private static Comparator<Ant> acomparator = new Comparator<Ant>() {
		public int compare(Ant o1, Ant o2) {
			  long val1 = o1.getScheduleAfter();
		      long val2 = o2.getScheduleAfter();
		      return (val1<val2 ? -1 : (val1==val2 ? 0 : 1));
		}
	};
		
	
	
	private static CyclicBarrier b;

	private static int threadcount;
	
	private static boolean cleanup() {
		synchronized(tbr) {
			synchronized(tba) {
				if ((tbr.size()==0) && (tba.size()==0)) return false;
				ants.removeAll(tbr);
				/*if (Parameters.RTT_MIGRATION) {
					Collections.sort(tba, acomparator);
				}*/
				tbr.clear();
				if (tba.size()==0) {
					return false;
				}
				ants.addAll(tba);
				tba.clear();
				return true;
			}
		}
	}
	

	private static void sortExecution() {
		if (ants.size() == 0) return;
		synchronized(ants) {
				if (Parameters.RTT_MIGRATION) {
					Collections.sort(ants, acomparator);
				}
		}
	}

	public static int getCount() {
		return ants.size();
	}

	public static void init(int count, int threads) {
		ants = new ArrayList<Ant>();
		tba = new ArrayList<Ant>(count);
		tbr = new ArrayList<Ant>(count);
		threadcount = threads;
		b = new CyclicBarrier(threads + 1);
		runners = new ArrayList<AntRunner>();
		initRunners();
	}

	public static void schedule(Ant ant) {
		synchronized(tba) {
			if (Parameters.PACKET_LOSS_PROB != 0 && Randomizer.pop() < Parameters.PACKET_LOSS_PROB) {
				Stats.loss();
			} else if (Parameters.PACKET_SLOW_PROB != 0 && Randomizer.pop() < Parameters.PACKET_SLOW_PROB) {
				Stats.slow();
				ant.setScheduleAfter(ant.getScheduleAfter() + Parameters.MS_PER_ITER);
			} else {
				tba.add(ant);
			}
		}
	}

	public static void initRunners() {
		if (threadcount == 0)
			return;
		runners = new ArrayList<AntRunner>(threadcount);
		for (int i = 0; i < threadcount; i++) {
			runners.add(new AntRunner(b));
		}
		for (AntRunner n : runners) {
			new Thread(n).start();
		}
	}

	private static void multithreadStep() {
		iterator = ants.iterator();
		for (AntRunner n : runners) {
			n.go();
		}
		try {
			b.await();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}
	}
	
	public static void step() {
		int asize = ants.size();
		if (asize + tba.size() == 0) return;
		if(cleanup())
			sortExecution();
		if ((ants.size() > 0) && (ants.get(0).getScheduleAfter() > NetworkSystem.getGlobalTimeIter())) return;
		if (Parameters.DO_MULTITHREAD && asize > 50) {
			multithreadStep();
			return;
		}
		iterator = ants.iterator();
		Ant next = AntSystem.getNextAnt();
		long netTime = NetworkSystem.getGlobalTimeIter();
		while (next != null) {
			if (next.getScheduleAfter() < netTime) {
				next.antstep();
			} else {
				break; // Ants coming here should not be executed
			}
			next = AntSystem.getNextAnt();
		}
	}

	public static void unschedule(Ant ant) {
		synchronized(tbr) {
			tbr.add(ant);
		}
	}

	public static Ant getNextAnt() {
		synchronized(iterator) {
			if (iterator.hasNext())
				return iterator.next();
			else
				return null;
		}
	}
		
	public static int getAntCount(Class<? extends Ant> type) {
		int count = 0;
		for (Ant a : ants) {
			if (a.getClass().equals(type))
				count++;
		}
		return count;
	}
	
	public static String getAntTypes() {
		Map<String,Integer> anc = new HashMap<String,Integer>();
		StringBuilder sr = new StringBuilder();
		for (Ant a : ants) {
			String type = a.getClass().getName();
			if (!anc.containsKey(type)) {
				anc.put(type, 1);
			} else {
				anc.put(type, anc.get(type)+1);
			}
		}
		sr.append(String.format("\tTotal: %d :: ", ants.size()));
		for (String k : anc.keySet()) {
			sr.append(String.format("%s (%d);", k, anc.get(k)));
		}
		return sr.toString();
	}

}
