/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package ants;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;

import pheromone.Pheromone;
import support.NetworkSystem;
import support.Stats;

public class BalancingAnt extends Ant {

	private enum States {
		MIN_SEARCH, MAX_SEARCH
	};

	private States state;
	private int hops;

	private int meanload = 0;
	private int totalload = 0;
	private int count = 0;
	private String previous = null;

	// Max and min node found
	private String max;
	private String min;

	// Max and min loads found
	private double max_val;
	private double min_val;

	private double load = 0;

	public BalancingAnt(String current) {
		super(current);
		state = States.MAX_SEARCH;
		max_val = -1;
		min_val = 10000;
	}

	@Override
	public void antstep() {
		if (!NetworkSystem.isNodeAlive(current_)) {
			// We are on a dead node
			kill();
			return;
		}
		current.acquire();
		Collection<String> targets = current.getNeighbors();
		final Pheromone stigma = current.getLoadManagementService().getStigma();
		current.release();

		if (targets.isEmpty())
			return;

		if ((meanload > 50) || (load > 50)) {
			if (previous != null) {
				stigma.fill(previous);
			}
		}

		load = NetworkSystem.getNode(current_).getLoadManagementService()
				.getLoad();

		totalload += load;
		count++;
		meanload = totalload / count;

		String t = null;

		if (state == States.MIN_SEARCH) {
			if (load < min_val) {
				min_val = load;
				min = current_;
			}
			// in MIN_SEARCH status follow lowest stigma trails
			t = Collections.min(targets, new Comparator<String>() {
				public int compare(String o1, String o2) {
					Double val1 = stigma.senseTrail(o1);
					Double val2 = stigma.senseTrail(o2);
					return val1.compareTo(val2);
				}
			});
		} else {
			if (load > max_val) {
				max_val = load;
				max = current_;
			}
			// in MAX_SEARCH status follow highest stigma trails
			t = Collections.min(targets, new Comparator<String>() {
				public int compare(String o1, String o2) {
					Double val1 = stigma.senseTrail(o1);
					Double val2 = stigma.senseTrail(o2);
					return val1.compareTo(val2);
				}
			});
		}

		if (hops++ == 9) {
			// Switch state
			state = (state == States.MAX_SEARCH) ? States.MIN_SEARCH
					: States.MAX_SEARCH;
			if ((max != null) && (min != null)) {
				// When switching status, if there is a valid highest and lowest
				// ask for balancing
				try {
					double maxLoad = NetworkSystem.getNode(max)
							.getLoadManagementService().getLoad();
					double minLoad = NetworkSystem.getNode(min)
							.getLoadManagementService().getLoad();
					double mean = (minLoad + maxLoad) / 2;
					NetworkSystem.getNode(max).getLoadManagementService()
							.setLoad(mean);
					NetworkSystem.getNode(min).getLoadManagementService()
							.setLoad(mean);
					Stats.balancing();
				} catch (Exception e) {
					// This will happen if a node dies
				}
				kill();
				return;
			}
			if (state == States.MAX_SEARCH) {
				max = null;
				min = null;
				max_val = -1;
				min_val = 1000;

			}
			hops = 0;
		}

		previous = current_;
		migrate(t);
	}

	@Override
	public void statmigration() {
		Stats.antMigration(BalancingAnt.class, 2);
	}

}
