/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package ants;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import pheromone.Pheromone;
import support.IBlatAntNode;
import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;
import alpha.Entry;

public class DiscoveryAnt extends Ant {

	private List<Entry> antInfos;
	private List<String> visited;
	private boolean migrationSuccess = true;
	private List<String> targets;
	private Entry local = null;

	public DiscoveryAnt(String current) {
		super(current);
		antInfos = new LinkedList<Entry>();
		visited = new LinkedList<String>();
		targets = new ArrayList<String>();
		local = new Entry(current);
	}

	public void antstep() {
		if (!NetworkSystem.isNodeAlive(current_)) {
			// We are on a dead node
			kill();
			return;
		}

		current.acquire();

		targets.clear();

		if (migrationSuccess) {
			((IBlatAntNode) current).updateInformation(antInfos);
			storeInfo();
		}

		List<String> ns = new LinkedList<String>();
		((IBlatAntNode) current).getInfosTo(local);
		local.getNeighbors(ns);
		for (String n : ns) {
			if (!visited.contains(n))
				// if (current.isActiveNeighbor(n))
				targets.add(new String(n));
		}

		if (targets.isEmpty()) {
			//visited.clear();
			targets.clear();
			local.getNeighbors(targets);
			// if (targets.size() == 0) {
			// System.out.println("Node "+ current_ +" is isolated
			// "+current.getNeighborsCount());
			// }
		}

		//save();
		double r = Randomizer.pop();
		if (r < Parameters.RANDOM_WALK_PROBABILITY) {
			// Random wandering
			int j = (int) Math.floor(Randomizer.pop() * targets.size());
			current.release();
			try {
				migrationSuccess = migrate(targets.get(j));
			} catch (Exception e) {
				// e.printStackTrace();
				// System.out.println("Error DAnt");
				// kill();
				return;
			}
		} else {
			// Follow lowest gamma concentration
			String t = null;
			final Pheromone gamma = current.getGamma();
			try {
				t = Collections.min(targets, new Comparator<String>() {
					public int compare(String o1, String o2) {
						Double val1 = gamma.senseTrail(o1);
						Double val2 = gamma.senseTrail(o2);
						return val1.compareTo(val2);
					}
				});
			} catch (Exception e) {
				// e.printStackTrace();
				// System.out.println("-------> "+targets);
			}
			current.release();
			migrationSuccess = migrate(t);
		}
	}

	protected boolean migrate(String target) {
		if (Parameters.PACKET_LOSS_PROB != 0 && Randomizer.pop() < Parameters.PACKET_LOSS_PROB) {
			Stats.loss();
			kill();
			return false;
		}
		if (NetworkSystem.isNodeAlive(target)) {
			setScheduleAfter(NetworkSystem.getRTT(current_, target));
			if (Parameters.PACKET_SLOW_PROB != 0 && Randomizer.pop() < Parameters.PACKET_SLOW_PROB) {
				Stats.slow();
				setScheduleAfter(getScheduleAfter() + Parameters.MS_PER_ITER);
			}
			current.acquire();
			((IBlatAntNode) current).getGamma().fill(target);
			((IBlatAntNode) current).tick();
			current.release();
			statmigration();
			INetworkNode t = NetworkSystem.getNode(target);
			t.acquire();
			t.getBeta().fill(current_);
			previous_ = current_;
			((IBlatAntNode) current).activateNeighbor(previous_); // Activate
																  // the
																  // neighbor,
																  // if
																  // necessary
			current_ = target;
			current = t;
			aging();
			((IBlatAntNode) t).tick();
			t.release();
			return true;
		} else {
			kill();
			return false;
		}
	}

	public void reset(String current) {
		super.reset(current);
		antInfos.clear();
		visited.clear();
		targets.clear();
	}

	private void storeInfo() {
		Entry ecopy = new Entry(current_);
		((IBlatAntNode) current).getInfosTo(ecopy);
		antInfos.add(ecopy);
		ecopy.setRemoteTime(new Long(current.getTime()));
		visited.add(ecopy.getId());
		while (antInfos.size() > Parameters.ANT_LEN) {
			antInfos.remove(0);
			visited.remove(0);
		}
	}

	@Override
	public void statmigration() {
		Stats.antMigration(DiscoveryAnt.class, antInfos.size());
		long total = 0;
		for (Entry e : antInfos) {
			total += e.getNeighborsCount();
		}
		Stats.nonTokens(total);
	}

}
