/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package ants;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import messages.Message;
import services.IResourceManagementService;
import support.IBlatAntNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;

public class QueryAnt extends Ant {

	private Message query;
	private String sourcetabu;
	private String previoustabu;
	private int ttl;
	private List<String> queryPath;

	public QueryAnt(String source, String previous, String target, Message query, int ttl, List<String> querypath) {
		super(target);
		this.query = query;
		this.ttl = ttl;
		this.sourcetabu = source;
		this.previoustabu = previous;
		this.queryPath = new ArrayList<String>();
		if (Parameters.QUERY_PATH_TRACKING && !(querypath == null))
			this.queryPath.addAll(querypath);
		setScheduleAfter(NetworkSystem.getRTT(previous, target));
	}

	@Override
	public void antstep() {
		
		if ((current == null) || !NetworkSystem.isNodeAlive(current_)) {
			// We are on a dead node
			kill();
			return;
		}
		
		current.acquire();
		
		// Manage beta pheromone
		if (this.previoustabu != null) {
			if (current instanceof IBlatAntNode) {
				((IBlatAntNode) current).activateNeighbor(previoustabu);
			}
			if (current.getBeta() != null) {
				current.getBeta().fill(new String(previoustabu));
			}
		}
		
		// Check the query

		// Check the query

		IResourceManagementService.ProcessingResult result = current.getResourceManagementService().processQuery(query,
				false);
				
		/* Check if the query matches */
		if (result.equals(IResourceManagementService.ProcessingResult.STOP)) {
			current.release();
			kill();
			return;
		}
		
		// Forward the query
		
		if (ttl > 0) {
			--ttl; // Count one hop
			List<String> candidates = new ArrayList<String>();
			candidates.addAll(current.getNeighbors());
			candidates.removeAll(queryPath);
			Collections.shuffle(candidates, Randomizer.getRandom());
			int limitffw = Parameters.SEARCH_LIMIT_FFW;
			int count = 0;
			
			if (Parameters.ALWAYS_FORWARD_LEAFS) {
				List<String> toSkip = new ArrayList<String>();
				for (String q : candidates) {
					if (current.getNeighborNeighborsSize(q) == 1) {
						// It's a leaf
						if ((!q.equals(sourcetabu)) && (!q.equals(current_)) && (!q.equals(previoustabu))) {
							// TTL set to 0 to prevent further forwarding if our info is wrong
							AntSystem.schedule(new QueryAnt(new String(sourcetabu), new String(current_), new String(q), query, 0, queryPath));
							Stats.messageHop(query, current.getIdentifier(), q);
							statmigration();
							if (current.getGamma() != null)
								current.getGamma().fill(q);
							toSkip.add(q);
						}
						
					}
				}
				candidates.removeAll(toSkip);
			}
		
			if (Parameters.QUERY_PATH_TRACKING)
				queryPath.add(current.getIdentifier());
			
			for (String q : candidates) {
				if ((limitffw != -1) && (count >= limitffw))
					break;
				if ((!q.equals(sourcetabu)) && (!q.equals(current_)) && (!q.equals(previoustabu))) {
					AntSystem.schedule(new QueryAnt(new String(sourcetabu), new String(current_), new String(q), query, ttl, queryPath));
					Stats.messageHop(query, current.getIdentifier(), q);
					statmigration();
					if (current.getGamma() != null)
						current.getGamma().fill(q);
					count++;
				}
			}			
		}
		
		current.release();
		kill();
		return;
	}
	
	@Override
	public void statmigration() {
		Stats.antMigration(QueryAnt.class, 1);
	}

}
