/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package network;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import support.Parameters;
import util.WeightedAStar;
import util.WeightedGraph;


public class RTTCacheNormal implements IRTT {
	
	private WeightedGraph wg;
	private WeightedAStar was;
	private HashMap<String,Long> weightTable;
	private FileWriter fw = null;
	private Set<String> nodes = new HashSet<String>();
	private List<String> paths = new ArrayList<String>();
	
	/* (non-Javadoc)
	 * @see network.IRTT#initRTT(boolean, java.lang.String)
	 */
	public void initRTT(boolean directed, String unetpath) {
		if (unetpath == null) return;

		if (!new File(unetpath+".cache").exists()) {
			try {
				wg = importYaml(directed, unetpath);
				was = new WeightedAStar(wg);
				fw = new FileWriter(unetpath+".cache");
				compute();
				if (Parameters.RTT_GENERATOR)
					paths.addAll(weightTable.keySet());
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			loadRTTCache(unetpath+".cache");
			if (Parameters.RTT_GENERATOR)
				paths.addAll(weightTable.keySet());
			return;
		}
	}
	
	/* (non-Javadoc)
	 * @see network.IRTT#getNodePool()
	 */
	public Set<String> getNodePool() {
		return nodes;
	}
	
	private void loadRTTCache(String cachefile) {
		weightTable = new HashMap<String,Long>();
		String line = null;
		BufferedReader br = null;

		System.out.println("Loading RTT cache from "+cachefile);
		
		try {
			br = new BufferedReader(new FileReader(cachefile));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}

		long total = 0;
		
		try {
			while ((line = br.readLine()) != null) {
				String[] data = line.split(":");
				String from = data[0].trim();
				nodes.add(from);
				String to = data[1].trim();
				nodes.add(to);
				Long weight = Long.parseLong(data[2].trim());
				String key = String.format("%s->%s", from,to);
				String key2 = String.format("%s->%s", to,from);
				weightTable.put(key, weight);
				weightTable.put(key2, weight);
				total += 2.0*weight;
			}
		} catch (NumberFormatException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		System.out.format("%d total NORMAL paths loaded, RTT Factor %d, average roundtrip %d ms\n", weightTable.size(), Parameters.MS_PER_ITER, total / weightTable.size());
		
		try {
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}			
	}

	/* (non-Javadoc)
	 * @see network.IRTT#getPathWeight(java.lang.String, java.lang.String)
	 */
	public long getPathWeight(String x, String y) {
		if ((x == null) || (y == null)) return 0;
		if (x.equals(y)) return 0;
		if (weightTable != null) {
			String key = String.format("%s->%s", x,y);
			if (Parameters.RTT_GENERATOR && !weightTable.containsKey(key)) {
				key = paths.get(Math.abs(key.hashCode()) % paths.size());
				return Math.round(weightTable.get(key) / Parameters.MS_PER_ITER_DOUBLE);
			} else
				return Math.round(weightTable.get(key) / Parameters.MS_PER_ITER_DOUBLE);
		}
		if (was != null)
			return Math.round(was.getDistance(x, y) / Parameters.MS_PER_ITER_DOUBLE);
		else
			return -1;
	}
	
	private void compute() {
		
		if (fw == null) return;
		
		System.out.println("Computing RTT... this may take a while");
		
		List<String> targets = new ArrayList<String>();
		targets.addAll(wg.vertexSet());
		
		for (String x : wg.vertexSet()) {
			nodes.add(x);
			for (String y : targets) {
				if (!x.equals(y)) {
					doPath(wg, was, x, y);
				}
			}
			targets.remove(x);
		}
	
		for (String x : wg.vertexSet())
			for (String y : wg.vertexSet()) {
				long weight = getPathWeight(x,y);
				try {
					if (!x.equals(y)) {
						fw.write(String.format("%s : %s : %d\n", x, y, weight));
						fw.write(String.format("%s : %s : %d\n", y, x, weight));
					}
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
		
		if (fw != null) {
			try {
				fw.flush();
				fw.close();
			} catch (IOException e) {
				e.printStackTrace();
			}

		}

	}
	
	private void doPath(WeightedGraph wg, WeightedAStar was, String source, String target) {
		List<String> path = was.getShortestPath(source, target);
		System.out.println(source+" -> "+target+" : "+path);
	}

	private WeightedGraph importYaml(boolean directed, String path) {
		WeightedGraph wg = new WeightedGraph(directed);
		String line = null;
		String currentNode = null;
		BufferedReader br = null;

		try {
			br = new BufferedReader(new FileReader(path));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}

		try {
			while ((line = br.readLine()) != null) {
				if (line.charAt(0) != ' ') {
					// Skip this
				} else {
					StringTokenizer st = new StringTokenizer(line, ":");
					String id = new String(st.nextToken().trim());
					if (id.equals("datastructure"))
						break;
					if (st.countTokens() > 0) {
						// Neighbor
						Double weight = null;
						if (st.hasMoreTokens()) {
							String wgs = new String(st.nextToken().trim());
							weight = Double.parseDouble(wgs);
							System.out.println("Weight for "+currentNode+"->"+id+" is "+weight.longValue());
						} else {
							System.out.println("No weight specified for "+currentNode+"->"+id);
							weight = 1.0;
						}
						wg.addVertex(id);
						wg.addEdge(currentNode, id, weight.longValue());
					} else {
						// New node
						wg.addVertex(id);
						currentNode = id;
					}
				}
			}
			br.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return wg;
	}
	
	/* (non-Javadoc)
	 * @see network.IRTT#mstoiter(long)
	 */
	public long mstoiter(long ms) {
		return Math.round(ms / Parameters.MS_PER_ITER_DOUBLE);
	}
	
	
	/* (non-Javadoc)
	 * @see network.IRTT#itertoms(long)
	 */
	public long itertoms(long iter) {
		return iter * Parameters.MS_PER_ITER;
	}

}
