/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
/**
 *  BlåtAnt Algorithm implementation
 *  
 *  Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 *  					Pervasive and Artificial Intelligence Research Group
 *  					Department of Informatics DIUF
 *  					University of Fribourg, Fribourg
 *  					Switzerland
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
package node;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import logging.StaticLogger;
import support.IBlatAntNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;
import algo.Algorithm;
import alpha.Alpha;
import alpha.Entry;
import ants.AntSystem;
import ants.ConstructionLinkAnt;
import ants.PingAnt;
import ants.UnlinkAnt;

public class BlatAntNode extends Node implements IBlatAntNode {

	protected Alpha alpha;
	private boolean active = false;
	private long retryConnectionIfIsolatedAt = 0;

	public BlatAntNode(String identifier, boolean active) {
		super(identifier);
		alpha = new Alpha(this);
		this.active = active;
		this.retryConnectionIfIsolatedAt = getTime() +  Parameters.CONNECTION_RETRY_INTERVAL_MS;
	}

	public boolean connect(String target) {
		if (isNeighbor(target))
			return true;
		if ((Parameters.MAX_OPTI_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS))
			return false;
		if (target.equals(id)) {
			System.out.println("Connecting to ourselves " + target);
			return false;
		}
		if (!alpha.isNeighbor(target)) {
			Stats.triedconnection(target, this.id);
			beta.fill(new String(target));
			gamma.fill(new String(target));
			alpha.onConnect(new String(target));
			return true;
		}
		return false;
	}

	public boolean disconnect(String target) {
		if (alpha.isNeighbor(target)) {
			removeNeighbor(target);
			gamma.clear(target);
			beta.clear(target);
			alpha.onDisconnect(target);
			return true;
		}
		return true;
	}

	public int getEstimatedDistance(String target) {
		return alpha.getEstimatedDistance(target);
	}

	public void getInfosTo(Entry e) {
		alpha.getInfosTo(e);
	}

	public Collection<String> getNeighbors() {
		return alpha.getNeighbors();
	}

	public int getNeighborsCount() {
		return alpha.getNeighborsCount();
	}

	public boolean isNeighbor(String id) {
		return alpha.isNeighbor(id);
	}

	public boolean isActiveNeighbor(String id) {
		return alpha.isActiveNeighbor(id);
	}

	private boolean removeDeadNeighbor(String target) {
		if (alpha.isNeighbor(target)) {
			if (alpha.getNeighborsCount() == 1) {
				// Initiate connection with a random hook in the network
				disconnect(target);
				initiateConnection(NetworkSystem.getQuasiRandomNode());
				return false;
			}

			List<String> hooks = alpha.getNeighbors(target);
			if (hooks.size() <= 1) {
				StaticLogger.log(target + " dead; " + id
						+ "; dead was leaf, do nothing");
				disconnect(target);
				Stats.deadRemove();
				return true;
			}

			if (Parameters.SPIDER_WEB_CRASH_RECOVERY) {
				AntSystem.schedule(new ConstructionLinkAnt(new String(id),
						new String(target), Parameters.EMERGENCY_TTL)); // Urgent!

				for (String h : hooks) {
					if (h.equals(id))
						continue;
					AntSystem.schedule(new ConstructionLinkAnt(new String(id),
							new String(h), Parameters.EMERGENCY_TTL)); // Urgent!
				}
			} else {
				// Circular connection

				List<String> toNotify = new ArrayList<String>();
				toNotify.addAll(hooks);
				Collections.sort(toNotify);

				for (int t = 0; t < toNotify.size() - 1; t++) {
					// Process all but the last
					String source = toNotify.get(t);
					String dest = toNotify.get(t + 1);
					AntSystem.schedule(new ConstructionLinkAnt(new String(
							source), new String(dest), new String(id),
							Parameters.LEAVE_LINK_TTL)); /* Urgent */
				}
			}

			disconnect(target);
			Stats.deadRemove();
			if (getNeighborsCount() == 0) {
				String h = NetworkSystem.getQuasiRandomNode();
				StaticLogger.log("No neighbors! Emergency reconnection with "
						+ h);
				AntSystem.schedule(new ConstructionLinkAnt(new String(id),
						new String(h), Parameters.USER_LINK_TTL)); // Not
																	// urgent,
																	// because
																	// we are
																	// alone
			}
		}
		return true;
	}

	private void initiateConnection(String hook) {
		AntSystem.schedule(new ConstructionLinkAnt(new String(id), new String(
				hook), Parameters.USER_LINK_TTL));
		retryConnectionIfIsolatedAt = getTime()
				+ Parameters.CONNECTION_RETRY_INTERVAL_MS;
	}

	private void removeNeighbor(String identifier) {
		alpha.removeNeighbor(identifier);
	}
	
	public void timestep() {
		alpha.step();
		/*
		 * Check if we have been for a too long time without neighbors, if yes
		 * re-start the connection procedure we a well known node
		 */
		if ((this.getNeighborsCount() == 0) || !active) {
			if (getTime() >= retryConnectionIfIsolatedAt) {
				// Next assumes that with that number of connections we are ok
				// if (this.getNeighborsCount() >=
				// Parameters.MAX_OPTI_CONNECTIONS) {
				// active = true;
				// } else {
				// System.out.println("Retry connection
				// ns="+this.getNeighborsCount()+" active?"+active);
				initiateConnection(NetworkSystem.getQuasiRandomNode());
				// }
			}
		}
		
		for (String n : gammaNodes) {
			if (isNeighbor(n))
				AntSystem.schedule(new PingAnt(new String(id), new String(n)));
		}
		gammaNodes.clear();
		
		
		for (String n : betaNodes) {
			if (!alpha.isActiveNeighbor(n)) { // Node never communicate with
												// us
				disconnect(n);
			} else if (alpha.isNeighbor(n) && alpha.isActiveNeighbor(n)) { // Node
																			// communicated
																			// with
																			// us
																			// but
																			// is
																			// now
																			// dead
				removeDeadNeighbor(n);
			}
		}
		betaNodes.clear();
		
		/* Generate ants */
		if (getTime() >= nextGeneration) {
			Algorithm.instance.generateAnt(id);
			nextGeneration += Parameters.IOTA_MS;
		}
	}
	
	private long nextGeneration = (long) (Randomizer.pop() * Parameters.IOTA_MS*2 + getTime());
	
	Set<String> gammaNodes = new HashSet<String>();
	Set<String> betaNodes = new HashSet<String>();

	public void iterstep() {
		

		// Future work: TODO: A node can, at random intervals and with random
		// probability try to connect to a node. This will recover network
		// partition
		// situations.

		/*
		 * If the gamma pheromone (out-pheromone) is low, not enough traffic did
		 * go on the corresponding link. So we need to send a ping.
		 */
		gammaNodes.addAll(gamma.decay());
		

		/*
		 * If the beta pheromone (in-pheromone) is low, not enough traffic
		 * arrived from a link. We initiate a recovery procedure.
		 */
		betaNodes.addAll(beta.decay());

	}

	public void tick() {
	}

	public void updateEntry(Entry e, int d) {
		alpha.updateEntry(e, d);
	}

	public void activateNeighbor(String neighbor) {
		alpha.activateNeighbor(neighbor);
	}

	public void updateInformation(List<Entry> info) {
		alpha.updateInformation(info);
	}

	public void updateNeighbors(String neighbor, Entry e) {
		alpha.updateNeighbors(neighbor, e);
	}

	public boolean userconnect(String target, int ttl) {

		// 0. If the node is not active, reject
		if (!active) {
			StaticLogger.logRejectedConnection(id, target, "Node not active");
			return false;
		}

		// 1. Check if we are not already neighbors
		if (isNeighbor(target)) {
			StaticLogger.logRejectedConnection(id, target,
					"Is already neighbor");
			return true;
		}

		// 2. Check if we are connecting to ourselves
		if (target.equals(id)) {
			StaticLogger.logRejectedConnection(id, target,
					"Connecting to ourselves");
			return false;
		}

		// 3. Check if we have still some free connections
		if ((Parameters.MAX_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= Parameters.MAX_CONNECTIONS)) {
			StaticLogger.logRejectedConnection(id, target, "Node is full");
			return false;
		}

		// 4. If the ant ttl is > 0, check if we are full of optimization
		// connectios
		if ((ttl > 0)
				&& (Parameters.MAX_OPTI_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS)) {
			StaticLogger.logRejectedConnection(id, target,
					"Node is almost full");
			return false;
		}

		// 5. If the node is not active, leave a free connection for the
		// activating connection
		if ((!active)
				&& (Parameters.MAX_OPTI_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= Parameters.MAX_OPTI_CONNECTIONS - 1)) {
			StaticLogger.logRejectedConnection(id, target,
					"Node is almost full and not yet active");
			return false;
		}

		// 6. Accept the connection
		beta.fill(new String(target)); // if beta is not filled here, if the
										// ant gets lost in its
		// way to the target (the source node requesting connection)
		// then we have a neighbor that is not tracked by a beta
		// pheromone, meaning that we lose one connection and we
		// are not able to notice it.
		gamma.fill(new String(target));
		alpha.onConnect(new String(target));
		return true;
	}

	/**
	 * Second step user connection (when OUR construction-link ant comes back)
	 */
	public boolean userconnect(String target) {

		// 1. Check if we are not already neighbors
		if (isNeighbor(target)) {
			StaticLogger.logRejectedConnection(id, target,
					"Is already neighbor");
			return true;
		}

		// 2. Check if the request does not come from ourselves
		if (target.equals(id)) {
			StaticLogger.logRejectedConnection(id, target, "Is same node");
			return false;
		}

		// 3. Check if we have enough free connections
		if ((Parameters.MAX_CONNECTIONS > 0)
				&& (alpha.getNeighborsCount() >= Parameters.MAX_CONNECTIONS)) {
			StaticLogger.logRejectedConnection(id, target, "Node is full");
			return false;
		}

		// 4. Accept the connection
		beta.fill(new String(target));
		gamma.fill(new String(target));
		alpha.onConnect(new String(target));
		active = true;

		activateNeighbor(target); // Since it answered (we are at step 2) it
									// can be considered as active
		AntSystem.schedule(new PingAnt(new String(id), new String(target)));

		return true;
	}


	public int getNeighborNeighborsSize(String n) {
		return alpha.getNeighbors(n).size();
	}

	public void cleanLeave() { // Create circular connection between neighbors
		if (getNeighborsCount() <= 1) // Only one neighbor, or none, do
										// nothing
			return;
		List<String> toNotify = new ArrayList<String>();
		toNotify.addAll(getNeighbors());
		Collections.sort(toNotify);
		if (toNotify.size() > 2) {
			String last = toNotify.get(toNotify.size() - 1);
			toNotify.add(0, last);
		}

		for (String ns : getNeighbors()) {
			AntSystem.schedule(new UnlinkAnt(id, new String(ns)));
		}

		for (int t = 0; t < toNotify.size() - 1; t++) {
			// Process all but the last
			String source = toNotify.get(t);
			String dest = toNotify.get(t + 1);
			AntSystem
					.schedule(new ConstructionLinkAnt(new String(source),
							new String(dest), new String(id),
							Parameters.LEAVE_LINK_TTL)); /* Urgent */
		}
	}

}
