/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import jsc.distributions.Poisson;
import messages.Message;
import messages.Query;
import node.Node;
import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import support.Stats;
import ants.AntSystem;
import ants.ISSProactiveQueryAnt;


/**
 * This class implements the CroissAnt 2nd level information system
 * The idea is to keep a small cache of nodes with similar profiles,
 * found with periodical light resource discovery or thorugh caching of
 * user queries results. 
 * The first are referred as "active" caching, whereas the second as "passive"
 * caching. Now we are interested in active caching only.
 * This cache is exchanged in a newscast like fashion. Resource discovery works
 * by sending queries with limited selective neighbor propagation (--limitffw): when
 * a node matching the request is found, the query is propagated to nodes in the cache
 * instead of on the base overlay.
 * 
 * @author attila
 *
 */

public class InformationSystemService implements IInformationSystemService {
	
	private Object cacheMutex = new Object();
	private InformationSystemEntry[] cache;
	private int cachesize = 0;
	private Node owner;
	private long nextMerge = 1;
	private long nextQuery = 1;
	private Map<String,Integer> pqueryttl;
	private List<String> pqueries;
	private List<String> toProcess = new ArrayList<String>();
	
	/* Insertion point */
	private int currentIndex = 0;
	private long localts = 0;

	private static Poisson poisson_ISS_PROACTIVE_QUERY_INTERVAL_MS = new Poisson(Math.max(Parameters.ISS_PROACTIVE_QUERY_INTERVAL_MS,1));
	private static Poisson poisson_ISS_MERGE_INTERVAL_MS = new Poisson(Math.max(Parameters.ISS_MERGE_INTERVAL_MS,1));
	private static Poisson poisson_ISS_MERGE_START_MS = new Poisson(Math.max(Parameters.ISS_MERGE_START_MS,1));
	private static Poisson poisson_ISS_PROACTIVE_QUERY_START_MS = new Poisson(Math.max(Parameters.ISS_PROACTIVE_QUERY_START_MS,1));
	private static Poisson poisson_ISS_PROACTIVE_QUERY_START_MS_FIRST = new Poisson(Math.max(Math.min(Parameters.ISS_PROACTIVE_QUERY_START_MS,Parameters.ISS_MERGE_START_MS),1));
	
	public InformationSystemService(Node owner, int cachesize) {
		if (!Parameters.ISS_ENABLED)
			return;
		this.cachesize = cachesize;
		this.cache = new InformationSystemEntry[cachesize];
		this.owner = owner;
		nextMerge = owner.getTime();
		nextQuery = owner.getTime();
		if (Parameters.ISS_MERGE_START_MS < 0) {
			this.nextMerge = -1;
		} else {
			this.nextMerge += poisson_ISS_MERGE_START_MS.random() + 1;
		}
		if (Parameters.ISS_PROACTIVE_QUERY_START_MS < 0) {
			this.nextQuery = -1;
		} else {
			if (Parameters.ISS_MERGE_START_MS < 0) {
				this.nextQuery += poisson_ISS_PROACTIVE_QUERY_START_MS.random() + 1;
			} else {
				this.nextQuery += poisson_ISS_PROACTIVE_QUERY_START_MS_FIRST.random() + 1;
			}
		}	
		
		pqueries = new LinkedList<String>();
		pqueryttl = new HashMap<String,Integer>();
	}
	
	public void cleanup() {
		while(pqueries.size() > Parameters.QUERY_CACHE_SIZE) {
			pqueryttl.remove(pqueries.remove(0));
		}
	}
	
	private String getRandomPeer() {
		synchronized(cache) {
			if (cacheIsEmpty()) return null;
			List<String> peers = new ArrayList<String>();
			for (InformationSystemEntry n : cache) {
				if (n != null) {
					peers.add(n.id);
				}
			}
			Collections.shuffle(peers);
			return peers.get(0);
		}
	}
	
	/*public void setCache(InformationSystemEntry[] cache) {
		this.cache = cache;
	}*/

	public InformationSystemEntry[] getCacheClone() {
		synchronized(cacheMutex) {
			return cache.clone();
		}
	}
	
	private void mergeData(IInformationSystemService remote) {
		InformationSystemEntry[] remoteCache = remote.getCacheClone();
		InformationSystemEntry[] localCache = getCacheClone();
		
		InformationSystemEntry[] temp = new InformationSystemEntry[2 * cachesize];
		
		int cost = remoteCache.length;

		//Stats.tokensTransmitted(2 * cachesize);

		for (int i = 0; i < cachesize; i++) {
			if (localCache[i] == null)
				continue;
			boolean done = false;
			for (int j = 0; j < cachesize * 2; j++) {
				if ((temp[j] != null) && (temp[j].id.equals(localCache[i].id))) {
					// There is already an entry
					temp[j].ts = Math.max(temp[j].ts, localCache[i].ts);
					done = true;
				}
			}
			if (done)
				continue;
			temp[i] = new InformationSystemEntry(localCache[i]);
		}

		for (int i = 0; i < cachesize; i++) {
			if (remoteCache[i] == null)
				continue;
			boolean done = false;
			for (int j = 0; j < cachesize * 2; j++) {
				if ((temp[j] != null) && (temp[j].id.equals(remoteCache[i].id))) {
					// There is already an entry
					temp[j].ts = Math.max(temp[j].ts, remoteCache[i].ts);
					done = true;
				}
			}
			if (done)
				continue;
			temp[cachesize + i] = new InformationSystemEntry(remoteCache[i]);
		}

		Arrays.sort(temp, new Comparator<InformationSystemEntry>() {

			public int compare(InformationSystemEntry arg0, InformationSystemEntry arg1) {
				if (arg0 == null)
					return 1;
				if (arg1 == null)
					return -1;
				return ((Long) arg1.ts).compareTo((Long) arg0.ts);
			}
		});

		for (int i = 1; i < cachesize; i++) {
			localCache[i] = temp[i - 1];
		}

		for (int i = 1; i < cachesize; i++) {
			remoteCache[i] = temp[i - 1];
		}

		long maxtime = Math.max(remote.getTime(), getTime());
		
		localCache[0] = new InformationSystemEntry();
		localCache[0].ts = new Long(maxtime);
		localCache[0].id = new String(remote.getIdentifier());
		remoteCache[0] = new InformationSystemEntry();
		remoteCache[0].ts = new Long(maxtime);
		remoteCache[0].id = new String(owner.getIdentifier());
		
		replaceCacheWith(localCache);
		remote.replaceCacheWith(remoteCache);
		
		Stats.issMerge(cost + remoteCache.length);
	}
	
	public void replaceCacheWith(InformationSystemEntry[] c) {
		synchronized(cacheMutex) {
			cache = c;
		}
	}

	public String getIdentifier() {
		return new String(owner.getIdentifier());
	}
	
	
	public void timestep() {
		if (!Parameters.ISS_ENABLED)
			return;
		
		if (!toProcess.isEmpty()) {
			for (String node : toProcess) {
				INetworkNode remoteNode = NetworkSystem.getNode(node);
				if (remoteNode != null) {
					//remoteNode.acquire();
					
					IInformationSystemService remoteService = remoteNode.getInformationSystemService();
					
					Stats.prohit();
					
					Stats.proReply();
					
					remoteService.pushPeer(new String(owner.getIdentifier()), getTime());
					
					//remoteNode.release();
				}
			}
			toProcess.clear();
		}
		
		if ((Parameters.ISS_MERGE_INTERVAL_MS > 0)
			&& (owner.getTime() >= nextMerge)) {
			
			String remote = getRandomPeer();
			if (remote != null) {
				
				INetworkNode remoteNode = NetworkSystem.getNode(remote);
				
				Stats.isspings();
				
				if (remoteNode != null) {
				
					//remoteNode.acquire();
					
					IInformationSystemService remoteService = remoteNode.getInformationSystemService();
		
					mergeData(remoteService);
					
					//remoteNode.release();
					
					nextMerge += poisson_ISS_MERGE_INTERVAL_MS .random() + 1;
					
				} else {
					
					removePeer(remote);
					
					nextMerge += 1;
					
				}
			} else {
				nextMerge += poisson_ISS_MERGE_INTERVAL_MS.random() + 1;
			}
			
		}
		
		if ((Parameters.ISS_PROACTIVE_QUERY_INTERVAL_MS > 0)
			&& (owner.getTime() >= nextQuery)) {
			Stats.prostarted();
			AntSystem.schedule(new ISSProactiveQueryAnt(new String(owner.getIdentifier()),
					null, new String(owner.getIdentifier()), new Query(owner.getIdentifier(), owner.getProfile()), Parameters.ISS_PROACTIVE_QUERY_TTL));
			nextQuery += poisson_ISS_PROACTIVE_QUERY_INTERVAL_MS.random() + 1;
		}
	}
	
	private void removePeer(String remote) {
		for (int i=0; i < cachesize; i++) {
			if (cache[i] != null)
				if (cache[i].id.equals(remote)) {
					cache[i] = null;
					break;
				}
		}
	}

	private boolean cacheIsEmpty() {
		synchronized(cache) {
			for (int j = 0; j < cachesize; j++) {
				if (cache[j] != null)
					return false;
			}
			return true;
		}
	}

	public boolean isInCache(String id) {
		return getCacheIdentifiers().contains(id);
	}
	
	public void pushPeer (String n, long ts) {
		synchronized(cache) {
			if (!isInCache(n) && !n.equals(owner.getIdentifier())) {
				cache[0] = new InformationSystemEntry();
				cache[0].ts = ts; //owner.getTime();
				cache[0].id = new String(n);
				currentIndex = (currentIndex + 1) % cachesize;
				if (localts <= ts) {
					localts = ts + 1;
				}
			}
		}
	}
	
	public synchronized boolean processQuery (Message query, int remainingTTL) {
		if (!pqueries.contains(query.getIdentifier())) {
			/* The query is new, and has to be matched */
			pqueries.add(query.getIdentifier());
			pqueryttl.put(query.getIdentifier(), new Integer(remainingTTL));
			
			if (owner.getProfile().isWithinSimilarityBoundsOf(query.getProfile())) {			
				// Postpone merge to avoid problem when two or more nodes
				// try to exchange that at the same time (found thanks to jstack :) )
				toProcess.add(query.getOrigin());
				
				return Parameters.NONSTOP_SEARCH; // Stop forwarding unless NON_STOP_SEARCH
			}

			return true;  // Continue forwarding 
		} else if (remainingTTL > pqueryttl.get(query.getIdentifier())) {
			/* The query is not new, but has an higher TTL than before, update it */
			pqueryttl.put(query.getIdentifier(), new Integer(remainingTTL));
			return true; // Continue forwarding 
		} else {
			/* The query is not new, and the TTL is lower than last seen, forward only if no limit*/
			return Parameters.NONLIMIT_SEARCH;
		}
	}
	
	public Collection<String> getCacheIdentifiers() {
		if (!Parameters.ISS_ENABLED)
			return null;
		synchronized(cacheMutex) {
			Collection<String> ns = new ArrayList<String>(cachesize);
			for (int i = 0; i < cachesize; i++) {
				InformationSystemEntry e = cache[i];
				if ((e != null) && (!e.id.equals(owner.getIdentifier())))
					ns.add(new String(e.id));
			}
			return ns;
		}
	}
	
	public long getMergeInterval() {
		return Parameters.ISS_MERGE_INTERVAL_MS;
	}

	public long getProactiveInterval() {
		return Parameters.ISS_PROACTIVE_QUERY_INTERVAL_MS;
	}
	
	public int getCacheSize() {
		int size = 0;
		for (int i = 0; i < cachesize; i++) {
			InformationSystemEntry e = cache[i];
			if ((e != null) && (!e.id.equals(owner.getIdentifier())))
				size++;
		}
		return size;
	}

	public long getTime() {
		return localts;
	}

}
