/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services;

import java.util.LinkedList;
import java.util.List;

import jsc.distributions.Poisson;
import messages.GroupMessage;
import messages.Message;
import services.dss.profile.ResourceProfile;
import support.INetworkNode;
import support.NetworkSystem;
import support.Parameters;
import support.Randomizer;
import support.Stats;
import ants.AntSystem;
import ants.ReplicantAnt;

public class ResourceManagementService implements IResourceManagementService {

	private int queryProcessed = 0;
	private INetworkNode node;
	private List<String> queries;
	private List<String> issqueries;
	private ProfileCacheMap profilecache = null;
	private static Poisson poisson_REPLICATION_ANT_INTERVAL_MS = new Poisson(Math.max(Parameters.REPLICATION_ANT_INTERVAL_MS,1));

	public ResourceManagementService(INetworkNode node) {
		this.node = node;
		queries = new LinkedList<String>();
		issqueries = new LinkedList<String>();
		
		if (Parameters.PROFILE_CACHE)
			profilecache = new ProfileCacheMap();
	}

	public void cleanup() {
		queryProcessed = 0;
		while(queries.size() > Parameters.QUERY_CACHE_SIZE) {
			queries.remove(0);
		}
		while(issqueries.size() > Parameters.QUERY_CACHE_SIZE) {
			issqueries.remove(0);
		}
	}

	public synchronized ProcessingResult processQuery(Message query, boolean iniss) {
		incrementQueryProcessed();
		if (iniss) {
			if (Parameters.ISS_DISJOINT_OVERLAYS) {
				if (!issqueries.contains(query.getIdentifier())) {
					/* The query is new, and has to be matched */
					issqueries.add(query.getIdentifier());
					if (matches(query)) {
						Stats.queryMatchOnISS(query, node.getIdentifier());
						return Parameters.NONSTOP_SEARCH ? IResourceManagementService.ProcessingResult.MATCH : IResourceManagementService.ProcessingResult.STOP ; // Stop forwarding unless NON_STOP_SEARCH
					}
					return IResourceManagementService.ProcessingResult.CONTINUE;  // Continue forwarding 
				} else {
					/* The query is not new, forward only if no limit*/
					return Parameters.NONLIMIT_SEARCH ? IResourceManagementService.ProcessingResult.CONTINUE : IResourceManagementService.ProcessingResult.STOP;
				}
			} else {
				if (!queries.contains(query.getIdentifier()) && !issqueries.contains(query.getIdentifier())) {
					/* The query is new, and has to be matched */
					issqueries.add(query.getIdentifier());
					if (matches(query)) {
						Stats.queryMatchOnISS(query, node.getIdentifier());
						return Parameters.NONSTOP_SEARCH ? IResourceManagementService.ProcessingResult.MATCH : IResourceManagementService.ProcessingResult.STOP ; // Stop forwarding unless NON_STOP_SEARCH
					}
					return IResourceManagementService.ProcessingResult.CONTINUE;  // Continue forwarding 
				} else {
					/* The query is not new, forward only if no limit*/
					return Parameters.NONLIMIT_SEARCH ? IResourceManagementService.ProcessingResult.CONTINUE : IResourceManagementService.ProcessingResult.STOP;
				}
			}
		} else {
			if (Parameters.ISS_DISJOINT_OVERLAYS) {
				if (!queries.contains(query.getIdentifier())) {
					/* The query is new, and has to be matched */
					queries.add(query.getIdentifier());
					if (matches(query)) {
						Stats.messageMatch(query, node.getIdentifier());
						if (!Parameters.ISS_ENABLED)
							return Parameters.NONSTOP_SEARCH ? IResourceManagementService.ProcessingResult.MATCH : IResourceManagementService.ProcessingResult.STOP ; // Stop forwarding unless NON_STOP_SEARCH
						else
							return IResourceManagementService.ProcessingResult.MATCH;
					}
					return IResourceManagementService.ProcessingResult.CONTINUE; // Continue forwarding 
				} else {
					/* The query is not new, forward only if no limit*/
					return Parameters.NONLIMIT_SEARCH ? IResourceManagementService.ProcessingResult.CONTINUE : IResourceManagementService.ProcessingResult.STOP;
				}
			} else {
				if (!queries.contains(query.getIdentifier()) && !issqueries.contains(query.getIdentifier())) {
					/* The query is new, and has to be matched */
					queries.add(query.getIdentifier());
					if (matches(query)) {
						Stats.messageMatch(query, node.getIdentifier());
						if (!Parameters.ISS_ENABLED)
							return Parameters.NONSTOP_SEARCH ? IResourceManagementService.ProcessingResult.MATCH : IResourceManagementService.ProcessingResult.STOP ; // Stop forwarding unless NON_STOP_SEARCH
						else
							return IResourceManagementService.ProcessingResult.MATCH;
					}
					return IResourceManagementService.ProcessingResult.CONTINUE; // Continue forwarding 
				} else {
					/* The query is not new, forward only if no limit*/
					return Parameters.NONLIMIT_SEARCH ? IResourceManagementService.ProcessingResult.CONTINUE : IResourceManagementService.ProcessingResult.STOP;
				}
			}
		}
	}

	public int getQueryProcessed() {
		return queryProcessed;
	}

	private boolean matches(Message query) {
		if (query instanceof GroupMessage)
			return true;
		else {
			if (Parameters.PROFILE_CACHE && (profilecache != null)) {
				synchronized(profilecache) {
					for (String k : profilecache.keySet()) {
						ResourceProfile p = profilecache.get(k);
						if (p.matches(query.getProfile())) {
							if (NetworkSystem.ping(k)) {
								Stats.replicantHit();
								Stats.messageMatch(query, k); // FIXME: This causes a mismatch with the cache hits (we should count separately when coming from cache)
							}
						}
					} 
				}
			}
			return node.getProfile().matches(query.getProfile());
		}	
	}
	
	public void publish(String node, ResourceProfile p) {
		if (profilecache != null) {
			synchronized(profilecache) {
				Stats.publish();
				profilecache.put(node, p);
			}
		}
	}

	private synchronized void incrementQueryProcessed() {
		queryProcessed++;
	}
	
	private long nextReplication = (long) (Randomizer.pop() * Parameters.REPLICATION_ANT_INTERVAL_MS) + 1;

	public void timestep() {
		if (!Parameters.PROFILE_CACHE)
			return;
		
		if ((Parameters.REPLICATION_ANT_INTERVAL_MS > 0)
			&& (node.getTime() >= nextReplication)) {
			Stats.replicantStarted();
			AntSystem.schedule(new ReplicantAnt(new String(node.getIdentifier()),
					null, new String(node.getIdentifier()), node.getProfile().copy(), Parameters.ONE_HOP_REPLICATION ? 1 :Parameters.D));
					
			nextReplication += poisson_REPLICATION_ANT_INTERVAL_MS.random() + 1;
		}
		
		if (Parameters.CACHE_TIME_MS > 0) {
			synchronized(profilecache) {
				profilecache.cleanup();
			}
		}
						
	}

}
