/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services.dss;

import java.util.HashMap;
import java.util.Map;

import services.dss.scheduler.IScheduler;

public class JobStatusManager implements IStatusManager {
	
	public static enum Status { UNKNOWN, SUBMITTED, SCHEDULED, READY, RUNNABLE, RUNNING, FINISHED, FAILED, REVOKED }
	
	private Map<String,Status> status;
	private Map<String,String> location;
	private Map<String,Double> timestamps;
	private IScheduler scheduler;
	
	public JobStatusManager(IScheduler s) {
		status = new HashMap<String,Status>();
		location = new HashMap<String,String>();
		timestamps = new HashMap<String,Double>();
		scheduler = s;
	}
	
	/* (non-Javadoc)
	 * @see services.dss.IStatusManager#setJobStatus(java.lang.String, services.dss.JobStatusManager.Status, java.lang.String)
	 */
	public void setJobStatus(String uuid, Status s, String node) {
		if (s == Status.REVOKED) {
			scheduler.forceUnschedule(uuid);
			removeJob(uuid);
			return;
		} else if ((s == Status.FAILED) || (s == Status.FINISHED)) {
			removeJob(uuid);
			return;
		}
		synchronized(status) {
			synchronized(location) {
				status.put(uuid, s);
				if ((getJobStatus(uuid) != Status.READY) && (getJobStatus(uuid) != Status.RUNNING)) {
					location.put(uuid, node);
				}
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see services.dss.IStatusManager#getJobStatus(java.lang.String)
	 */
	public Status getJobStatus(String uuid) {
		synchronized(status) {
			if (!status.containsKey(uuid)) return Status.UNKNOWN;
			else return status.get(uuid);
		}
	}
	
	
	/* (non-Javadoc)
	 * @see services.dss.IStatusManager#getJobLocation(java.lang.String)
	 */
	public String getJobLocation(String uuid) {
		synchronized(location) {
			if (!location.containsKey(uuid)) return null;
			else return location.get(uuid);
		}
	}
	
	/* (non-Javadoc)
	 * @see services.dss.IStatusManager#removeJob(java.lang.String)
	 */
	public void removeJob(String uuid) {
		synchronized(status) {
			synchronized(location) {
				synchronized(timestamps) {
					status.remove(uuid);
					location.remove(uuid);
					timestamps.remove(uuid);
				}
			}
		}
	}

	public boolean updateTimestamp(String uuid, Double ts) {
		synchronized(timestamps) {
			if (!timestamps.containsKey(uuid)) return false;
			Double last = timestamps.get(uuid);
			if (ts < last) return false;
			timestamps.put(uuid, ts);
			return true;
		}
	}

	public void addJob(String uuid) {
		synchronized(status) {
			synchronized(timestamps) {
				status.put(uuid, Status.SUBMITTED);
				timestamps.put(uuid, -1.0);
			}
		}
	}


}
