/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services.dss.scheduler;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import services.dss.DSSStats;
import services.dss.Messaging;
import services.dss.JobStatusManager.Status;
import services.dss.profile.ResourceProfile;
import support.INetworkNode;
import support.Parameters;
import support.Randomizer;

public class ARScheduler implements IScheduler {
	
	// Advance Reservation Scheduler
	
	/**
	 * The cost function is:
	 * 
	 * cost (job) = the global cost is the sum of all delays
	 * 				if the global cost is = 0 (i.e. no job will delay) the cost is -(sum of free time between jobs)
	 * 				else the cost is the sum of delays 
	 */
	
	private Long currentTime = 0l;
	private INetworkNode d;
	
	private static Comparator<ResourceProfile> arComparator = new Comparator<ResourceProfile>() {

		public int compare(ResourceProfile arg0,
				ResourceProfile arg1) {
			if (arg0.isLocked() || arg0.isRunning()) return -1;
			return arg0.reservationStart.compareTo(arg1.reservationStart);
		}
		
	};
	
	private static Comparator<ResourceProfile> delayComparator = new Comparator<ResourceProfile>() {

		public int compare(ResourceProfile o1, ResourceProfile o2) {
			return o1.tempDelay.compareTo(o2.tempDelay);
		}
		
	};

	private List<ResourceProfile> schedule = new LinkedList<ResourceProfile>();
	private double cpuspeed = 1.0;
	
	private List<ResourceProfile> getTempSchedule(List<ResourceProfile> actual, ResourceProfile newReservation) {
		LinkedList<ResourceProfile> temp = new LinkedList<ResourceProfile>();
		temp.addAll(actual);
		temp.add(newReservation);
		Collections.sort(temp, arComparator);
		return temp;
	}
	
	public ARScheduler(INetworkNode d, double cpuspeed) {
		this.cpuspeed = cpuspeed;
		this.d = d;
	}
	
	private Long getCost(List<ResourceProfile> schedule) {
		Long delayTime = 0l;
		Long tempDelay = 0l;
		
		for (ResourceProfile r : schedule) {
			tempDelay = getDelay(schedule, r);
			if (tempDelay > 0)
				delayTime += tempDelay;
		}
		
		if (delayTime > 0)
			return delayTime;
		else
			return -1* getFreeTime(schedule);
	}
	
	private long getFreeTime(List<ResourceProfile> schedule) {
		// This does not check that the schedule does not overlap!!!
		long freeTime = 0l;
		long prevEnd = currentTime;
		
		for (ResourceProfile r : schedule) {
			if (r.isRunning()) {
				prevEnd = r.reservationEnd;
				continue;
			}
			freeTime += r.reservationStart - prevEnd;
			prevEnd = r.reservationEnd;
		}
		return freeTime;
	}

	private Long getETC(List<ResourceProfile> temp, ResourceProfile j) {
		long delay = getDelay(temp, j);
		
		return delay + j.reservationEnd;	
	}
	
	private Long getDelay(List<ResourceProfile> temp, ResourceProfile j) {
		Long delayTime = 0l;
		Long prevEnd = currentTime;
		
		for (ResourceProfile r : temp) {
			if (r.isRunning()) {
				delayTime = (r.startedAt - r.reservationStart);
			} else {
				long diff = r.reservationStart - prevEnd;
				if (diff < 0) {
					delayTime = Math.abs(diff);
				} else {
					delayTime = 0l;
				}
			}
			prevEnd = r.reservationEnd + delayTime;
			if (r == j) break;
		}
		
		return delayTime;	
	}
	
	

	public ResourceProfile forceUnschedule(String jobid) { 
		synchronized (schedule) {
			ResourceProfile j = getProfile(jobid);
			if (j != null)
				if (schedule.remove(j)) {
					return j;
				}
		}
		return null;
	}

	public double getCPUSpeed() {
		return cpuspeed ;
	}

	public Long getEstimatedCompletionTime(ResourceProfile j) {
		List<ResourceProfile> temp = new ArrayList<ResourceProfile>();
		synchronized(schedule) {
			temp.addAll(schedule);
			if (!temp.contains(j))
				temp.add(j);
			Collections.sort(temp, arComparator);
		}
		return getETC(temp, j);
	}

	public List<String> getJobs() {
		synchronized(schedule) {
			return getJobsId(schedule);
		}
	}
	
	private List<String> getJobsId(List<ResourceProfile> schedule2) {
		List<String> jid = new ArrayList<String>();
		synchronized (schedule2) {
			for (ResourceProfile f : schedule2) {
				jid.add(f.id);
			}
		}
		return jid;
	}

	public List<String> getJobsForInform() {
		List<ResourceProfile> temp = new ArrayList<ResourceProfile>();
		synchronized (schedule) {
			updateDelays();
			temp.addAll(schedule);
		}
		Collections.sort(temp, delayComparator);
		return getJobsId(temp);
	}

	private void updateDelays() {
		for (ResourceProfile f : schedule) {
			f.tempDelay = getDelay(schedule, f);
		}
	}

	public ResourceProfile getProfile(String jobid) {
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.id.equals(jobid))
					return f;
			}
		}
		return null;
	}

	public Long getScheduleLength() {
		if (schedule.isEmpty()) return 0l;
		return schedule.get(schedule.size()-1).reservationEnd;
	}

	public int getScheduleSize() {
		synchronized (schedule) {
			return schedule.size();
		}
	}

	public String getSchedulerName() {
		return "AR";
	}

	public Long getScore(ResourceProfile j) {
		List<ResourceProfile> temp = new ArrayList<ResourceProfile>();
		synchronized(schedule) {
			temp.addAll(schedule);
			if (!temp.contains(j))
				temp.add(j);
			Collections.sort(temp, arComparator);
		}
		return getCost(temp);
	}

	public boolean isScheduledJob(String jobid) {
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.id.equals(jobid))
					return true;
			}
		}
		return false;
	}

	public void progress() {
		currentTime = d.getTime();
		synchronized (schedule) {
			if (schedule.size() == 0)
				return;
			updateDelays();
			ResourceProfile running = schedule.get(0);
			if (!running.isRunning()) {
				if (running.reservationStart <= currentTime) {
					running.startedAt = currentTime;
					running.startDelay = currentTime - running.reservationStart;
					running.setRunning();
					running.tempDelay = 0l;
					if (running.actualRunningTime != -1) {
						running.remainingTime = (long) (running.actualRunningTime / cpuspeed);
						running.drift = 0l;
					} else {
						running.remainingTime = running.getEstimatedRunningTime(cpuspeed);
						running.drift = (long) ((Randomizer.pop() - 0.5) * running.getEstimatedRunningTime(cpuspeed) * (Parameters.DSS_DRIFT * 2)); // Plus minus 10%
						if (Parameters.BAD_DRIFT_ONLY)
							running.drift = Math.abs(running.drift);
					}
					DSSStats.started(running.id);
					
					/* Notify status to initiator */
					double ts = Double.parseDouble(String.format("%d.%d", running.assigns, d.getTime()));
					Messaging.sendStatus(d, running.id, running.initiator, Status.RUNNING, ts);
					/* End notify to initiator */
				}
			} else {
				running.remainingTime = running.getEstimatedRunningTime(cpuspeed) - (currentTime - running.startedAt);
				if ((running.remainingTime + running.drift) <= 0l) {
					running.remainingTime = 0l;
					running.finishedAt = currentTime;
					schedule.remove(0);
					DSSStats.finished(running, cpuspeed);
					//DSSStats.deadline(currentTime, running.deadline);
					DSSStats.reservation(currentTime, running.reservationStart, running.startedAt, running.reservationEnd, currentTime);
					
					/* Notify status to initiator */
					double ts = Double.parseDouble(String.format("%d.%d", running.assigns, d.getTime()));
					Messaging.sendStatus(d, running.id, running.initiator, Status.FINISHED, ts);
					/* End notify to initiator */
				}		
			}
		}
	}

	public void schedule(ResourceProfile j) {
		synchronized (schedule) {
			schedule = getTempSchedule(schedule, j);
		}
	}

	public ResourceProfile unschedule(String jobid) {
		synchronized (schedule) {
			ResourceProfile j = getProfile(jobid);
			if ((j != null) && (!j.isLocked()))
				if (schedule.remove(j)) {
					return j;
				}
		}
		return null;
	}

	public boolean isDeadlineScheduling() {
		return false;
	}

	public boolean isReservationScheduling() {
		return true;
	}

}
