/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
/**
 *  This file is part of
 *  
 *  Solenopsis 2: A Framework for the Development of Ant Algorithms
 *  
 *  Copyright (C) 2006-2009 Amos Brocco <amos.brocco@unifr.ch>
 *  						Pervasive and Artificial Intelligence Research Group
 *  						Department of Informatics DIUF
 *  						University of Fribourg, Fribourg
 *  						Switzerland
 *
 *  Solenopsis 2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  Solenopsis 2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with Solenopsis 2.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */
package services.dss.scheduler;

import java.util.ArrayList;
import java.util.List;

import services.dss.DSSStats;
import services.dss.Messaging;
import services.dss.JobStatusManager.Status;
import services.dss.profile.ResourceProfile;
import support.INetworkNode;
import support.Parameters;
import support.Randomizer;

public class FIFOScheduler implements IScheduler {

	long currentTime = 0l;

	List<ResourceProfile> schedule = new ArrayList<ResourceProfile>();

	private double cpuspeed = 1;
	private INetworkNode d;
	
	public FIFOScheduler(INetworkNode d, double cpuspeed) {
		this.cpuspeed  = cpuspeed;
		this.d = d;
	}
	
	public double getCPUSpeed() {
		return cpuspeed;
	}

	public Long getEstimatedCompletionTime(ResourceProfile j) {
		long waitingTime = 0;
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (!f.isRunning())
					waitingTime += f.getEstimatedRunningTime(cpuspeed);
				else
					waitingTime += Math.max(0,f.getEstimatedRunningTime(cpuspeed)
							- (currentTime - f.startedAt));
				if (f.equals(j))
					break;
					
			}
		}
		return waitingTime + j.getEstimatedRunningTime(cpuspeed);
	}

	public List<String> getJobs() {
		List<String> jid = new ArrayList<String>();
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				jid.add(f.id);
			}
		}
		return jid;
	}

	public ResourceProfile getProfile(String jobid) {
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.id.equals(jobid))
					return f;
			}
		}
		return null;
	}

	public boolean isScheduledJob(String jobid) {
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.id.equals(jobid))
					return true;
			}
		}
		return false;
	}

	public synchronized void progress() {
		currentTime = d.getTime();
		synchronized (schedule) {
			if (schedule.size() == 0)
				return;
			ResourceProfile running = schedule.get(0);
			if (!running.isRunning()) {
				running.startedAt = currentTime;
				//System.out.println("Starting job "+running.id);
				running.setRunning();
				
				if (running.actualRunningTime != -1) {
					running.remainingTime = (long) (running.actualRunningTime / cpuspeed);
					running.drift = 0l;
				} else {
					running.remainingTime = running.getEstimatedRunningTime(cpuspeed);
					running.drift = (long) ((Randomizer.pop() - 0.5) * running.getEstimatedRunningTime(cpuspeed) * (Parameters.DSS_DRIFT * 2)); // Plus minus 10%
					if (Parameters.BAD_DRIFT_ONLY)
						running.drift = Math.abs(running.drift);
				}
				DSSStats.started(running.id);
				
				/* Notify status to initiator */
				double ts = Double.parseDouble(String.format("%d.%d", running.assigns, d.getTime()));
				Messaging.sendStatus(d, running.id, running.initiator, Status.RUNNING, ts);
				/* End notify to initiator */
			} else {
				running.remainingTime = running.getEstimatedRunningTime(cpuspeed) - (currentTime - running.startedAt);
				if ((running.remainingTime + running.drift) <= 0l) {
					running.remainingTime = 0l;
					running.finishedAt = currentTime;
					schedule.remove(0);
					//System.out.println("Finished job "+running.id);
					DSSStats.finished(running, cpuspeed);
					
					/* Notify status to initiator */
					double ts = Double.parseDouble(String.format("%d.%d", running.assigns, d.getTime()));
					Messaging.sendStatus(d, running.id, running.initiator, Status.FINISHED, ts);
					/* End notify to initiator */
				}		
			}
		}
	}

	public void schedule(ResourceProfile j) {
		synchronized (schedule) {
			schedule.add(j); // New jobs join at the end
		}
	}

	public ResourceProfile unschedule(String jobid) {
		synchronized (schedule) {
			ResourceProfile j = getProfile(jobid);
			if ((j != null) && (!j.isLocked()))
				if (schedule.remove(j)) {
					return j;
				}
		}
		return null;
	}
	
	public ResourceProfile forceUnschedule(String jobid) {
		synchronized (schedule) {
			ResourceProfile j = getProfile(jobid);
			if (j != null)
				if (schedule.remove(j)) {
					return j;
				}
		}
		return null;
	}
	
	public Long getScheduleLength() {
		long length = 0l;
		synchronized (schedule) {
			for (ResourceProfile f : schedule) {
				if (f.isRunning())
					length += f.remainingTime;
				else
					length += f.getEstimatedRunningTime(cpuspeed);
			}
		}
		return length;
	}

	public int getScheduleSize() {
		synchronized (schedule) {
			return schedule.size();
		}
	}

	public String getSchedulerName() {
		return "FIFO";
	}

	public Long getScore(ResourceProfile j) {
		return this.getEstimatedCompletionTime(j);
	}
	
	public List<String> getJobsForInform() {
		return getJobs();
	}
	
	public boolean isDeadlineScheduling() {
		return false;
	}

	public boolean isReservationScheduling() {
		return false;
	}

}
