/*******************************************************************************
 * BlåtAnt Algorithm Evaluation Platform
 * 
 * Copyright (C) 2008  Amos Brocco <amos.brocco@unifr.ch>
 * 					Pervasive and Artificial Intelligence Research Group
 * 					Department of Informatics DIUF
 * 					University of Fribourg, Fribourg
 * 					Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 ******************************************************************************/
package services.dss.scheduler.pool;

import java.util.ArrayList;
import java.util.List;

import services.dss.DSSStats;
import services.dss.IStatusManager;
import services.dss.Messaging;
import services.dss.JobStatusManager.Status;
import services.dss.profile.ResourceProfile;
import support.INetworkNode;
import support.Parameters;
import support.Randomizer;

public class Pool {
	
	private String id;
	private ResourceProfile base;
	private List<String> jobs = new ArrayList<String>();
	
	/* Simulation data */
	private Double lowestCPUspeed = 2.0;
	private INetworkNode d;
	private IStatusManager jsm;
	
	public Pool(INetworkNode d, String uuid, ResourceProfile base) {
		this.base = base;
		this.id = uuid;
		this.d = d;
		this.jsm = d.getDSSService().getJobStatusManager();
	}
	
	public synchronized void addJob(String uuid) {
		jobs.add(uuid);
	}
	
	public synchronized void onJobReady(String uuid, String node, Double cpuspeed) {
		if (!jobs.contains(uuid)) return;
		setLowestCPUSpeed(cpuspeed);
		if (isPoolReady()) {
			/* Simulation values */
			long remainingTime = base.getEstimatedRunningTime(lowestCPUspeed);
			long drift = (long) ((Randomizer.pop() - 0.5) * base.getEstimatedRunningTime(cpuspeed) * (Parameters.DSS_DRIFT * 2)); // Plus minus 10%
			if (Parameters.BAD_DRIFT_ONLY)
				drift = Math.abs(drift);
			/* End simulation values */
			for (String job : jobs) {
				Messaging.sendStatus(d, job, jsm.getJobLocation(job), Status.RUNNABLE, new Double(d.getTime()), remainingTime, drift);
			}
		}
	}
	
	private boolean isPoolReady() {
		for (String j : jobs) {
			if (jsm.getJobStatus(j) != Status.READY)
				return false;
		}
		return true;
	}

	private void setLowestCPUSpeed(Double cpuspeed) {
		if (cpuspeed < lowestCPUspeed) {
			lowestCPUspeed = cpuspeed;
		}
	}
	
	public String getUUID() {
		return id;
	}
	
	public synchronized boolean hasJob(String uuid) {
		return jobs.contains(uuid);
	}

	public synchronized void revokeAll() {
		if (jobs.isEmpty()) return;
		DSSStats.revokePool(id, getSize());
		for (String job : jobs) {
			Messaging.sendStatus(d, job, jsm.getJobLocation(job), Status.REVOKED, new Double(d.getTime()));
		}
		jobs.clear();
	}

	public int getSize() {
		return jobs.size();
	}

	public List<String> getJobs() {
		return jobs;
	}
	
	
	

}
